//mini net :
class snn_mv_layer{
//the mv version has no nodes 
//the network receives the matrix of features directly so there is no need for an input layer
//we have a vector of outputs and their derivatives 
         public:
  matrix output,derivatives;//[nodes,1]
  matrix weights,weight_adjustments,weightsT;//a matrix of weights and adjustments to those weights 
         /*
         the weights matrix looks like this :
         each row represents a node in this layer 
         and each column (apart from the bias) represents a node in the previous layer
         */
         //the error on the nodes
  matrix error;//[nodes,1]
         snn_mv_layer(void){reset();}
        ~snn_mv_layer(void){reset();}
    void reset(){
         error.Resize(0,0);
         output.Resize(0,0);
         derivatives.Resize(0,0);
         weights.Resize(0,0,0);
         weight_adjustments.Resize(0,0);
         weightsT.Resize(0,0);
         } 
         /* setup function only one since there are no input layers */
    void setup(int nodes_total,//nodes of this layer
               int previous_layer_nodes_total,//
               bool randomize_weights=true,
               double weights_min=-1.0,
               double weights_max=1.0){
         error.Resize(nodes_total+1,1);//
         output.Resize(nodes_total+1,1);//+1 for the bias as its multiplied with the next layer matrix
         derivatives.Resize(nodes_total+1,1);
         /*
         weights :
         -how many nodes does this layer have ? thats the rows 
         -how many nodes did the previous layer have ? thats the columns + 1 (for the bias)
         */
         weights.Init(nodes_total+1,previous_layer_nodes_total+1);
         weight_adjustments.Init(nodes_total+1,previous_layer_nodes_total+1);
         weightsT.Init(nodes_total+1,previous_layer_nodes_total+1);
         if(randomize_weights){
           double range=weights_max-weights_min;
           for(int r=0;r<nodes_total;r++){
              for(int c=0;c<=previous_layer_nodes_total;c++){
                 weight_adjustments[r][c]=0.0;
                 weights[r][c]=(((double)MathRand())/32767.0)*range+weights_min;
                 }
              }
           //last row 
             for(int c=0;c<previous_layer_nodes_total;c++){
              weight_adjustments[nodes_total][c]=0.0;
              weights[nodes_total][c]=0.0;
              }
              weight_adjustments[nodes_total][previous_layer_nodes_total]=0.0;
              weights[nodes_total][previous_layer_nodes_total]=1.0;
           weightsT=weights.Transpose();
           }
         }
    void feed_forward(matrix &inputs_or_previous_output){
         output=weights.MatMul(inputs_or_previous_output);
         //activation
           
           output.Derivative(derivatives,AF_SIGMOID);
           output.Activation(output,AF_SIGMOID);
         } 
         //for the first layer no propagate 
    void back_propagation(matrix &inputs_of_sample){
         error*=derivatives;
         /* now each error must be multiplied by all the inputs 
            to get the weight adjustment */
         weight_adjustments+=error.MatMul(inputs_of_sample.Transpose());        
         }
         //fo the middle layers
    void back_propagation(snn_mv_layer &previous_layer){
         error*=derivatives;
         previous_layer.error=weightsT.MatMul(error);
         weight_adjustments+=error.MatMul(previous_layer.output.Transpose()); 
         }
    void adjust(double learning_rate){
         weights-=learning_rate*weight_adjustments;
         weightsT=weights.Transpose();
         reset_weight_adjustments();
         }
    void reset_weight_adjustments(){
         weight_adjustments.Fill(0.0);
         }
  double output_layer_calculate_loss(double &correct_outputs[]){
         //the correct outputs are the same # as the nodes of the output layer
           //the loss function we'll use is the quadratic 
           //its :  where i is the total output nodes
             //this is the value we'll return to the network , not the backpropagation value
               //you could use this to control the learning rate for instance
                 double loss=0.0;
                 for(int i=0;i<ArraySize(correct_outputs);i++){
                 loss+=MathPow((output[i][0]-correct_outputs[i]),2.0);
                 }
                 loss/=2.0;
         //and the error per output node is the derivative of 1/2(f-c)^2
         //which is : f-c
           for(int i=0;i<ArraySize(correct_outputs);i++){
              //so the error post activation for output node i is : 
                error[i][0]=output[i][0]-correct_outputs[i];
              }
         return(loss);
         }          
};

class snn_mv_net{
             public:
snn_mv_layer layers[];
         int samples_calculated;//for the batch
      double total_loss;
             /* a features + 1 by 1 matrix */
      matrix features;
             snn_mv_net(void){reset();}
            ~snn_mv_net(void){reset();}
        void reset(){
             ArrayFree(layers);
             samples_calculated=0;
             total_loss=0.0;
             }
        void setup(int features_total){
             features.Init(features_total+1,1);
             features.Fill(1.0);
             }
        void add_layer(int nodes_total,//nodes of this layer
                       int previous_layer_nodes_total,//
                       bool randomize_weights=true,
                       double weights_min=-1.0,
                       double weights_max=1.0){
             int ns=ArraySize(layers)+1;
             ArrayResize(layers,ns,0);
             layers[ns-1].setup(nodes_total,previous_layer_nodes_total,randomize_weights,weights_min,weights_max);
             }
        void feed_forward(double &_features[]){
             for(int i=0;i<ArraySize(_features);i++){features[i][0]=_features[i];}
             //bias stays 1.0
               //first layer 
                 layers[0].feed_forward(features);
                 for(int i=1;i<ArraySize(layers);i++){
                    layers[i].feed_forward(layers[i-1].output);
                    }
             }        
        void back_propagation_same_features(){
             for(int i=ArraySize(layers)-1;i>=1;i--){
                layers[i].back_propagation(layers[i-1]);
                }
                layers[0].back_propagation(features);
             }
        void back_propagation(double &_features[]){
             for(int i=0;i<ArraySize(_features);i++){features[i][0]=_features[i];}
             back_propagation_same_features();
             }
        void adjust(double learning_rate){
             for(int i=0;i<ArraySize(layers);i++){
                layers[i].adjust(learning_rate);
                }
             }
        void calculate_loss(double &sample_correct_outputs[]){
             samples_calculated++;
             //calculate the loss at the output layer and also add to the total loss
             total_loss+=layers[ArraySize(layers)-1].output_layer_calculate_loss(sample_correct_outputs);
             }             
      double get_max_loss_per_sample(){
             double max_loss_per_sample=(layers[ArraySize(layers)-1].output.Rows()-1)/2.0;
             return(max_loss_per_sample);
             }
      double get_current_loss_per_sample(){
             double current_loss_per_sample=total_loss/((double)samples_calculated);
             return(current_loss_per_sample);
             }
      double get_output(int ix){
             return(layers[ArraySize(layers)-1].output[ix][0]);
             }
        void reset_loss(){
             total_loss=0.0;
             samples_calculated=0;
             }
};
